@@ -6,19 +6,19 @@
  * Copyright 2001, 2002 Ximian, Inc.
  *
  * This library is free software; you can redistribute it and/or
- * modify it under the terms of the GNU Library General Public
+ * modify it under the terms of the GNU Lesser General Public
  * License as published by the Free Software Foundation; either
- * version 2 of the License, or (at your option) any later version.
+ * version 2.1 of the License, or (at your option) any later version.
  *
  * This library is distributed in the hope that it will be useful,
  * but WITHOUT ANY WARRANTY; without even the implied warranty of
  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
- * Library General Public License for more details.
+ * Lesser General Public License for more details.
  *
- * You should have received a copy of the GNU Library General Public
+ * You should have received a copy of the GNU Lesser General Public
  * License along with this library; if not, write to the
- * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
- * Boston, MA 02111-1307, USA.
+ * Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
+ * Boston, MA 02110-1301, USA.
  */
 
 #ifndef _ATSPI_H
@@ -26,12 +26,11 @@
 
 #include "glib.h"
 
-#include "atspi-constants.h"
-#include "atspi-types.h"
 #include "atspi-accessible.h"
 #include "atspi-action.h"
 #include "atspi-collection.h"
 #include "atspi-component.h"
+#include "atspi-constants.h"
 #include "atspi-device-listener.h"
 #include "atspi-document.h"
 #include "atspi-editabletext.h"
@@ -45,9 +44,13 @@
 #include "atspi-relation.h"
 #include "atspi-selection.h"
 #include "atspi-stateset.h"
+#include "atspi-table-cell.h"
 #include "atspi-table.h"
 #include "atspi-text.h"
+#include "atspi-types.h"
 #include "atspi-value.h"
 
 #include "atspi-gmain.h"
+
+#include "atspi-enum-types.h"
 #endif
